/*
 Copyright 2012, S.S. Papadopulos & Assoc. Inc.

    This file is part of GENIE.

    The GENIE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    GENIE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with GENIE.  If not, see<http://www.gnu.org/licenses/>.
*/

// cmuffels, Genie version 1, 2012

#include <iostream>

#include "runmanager.h"

using namespace std;

//****************************************************************************
void RUN_MANAGER::initialize()
//****************************************************************************
//CTM Apr 2011

/*
      Initialize run manager
*/

{

  //threads=new list<RECEIVOR*>;
  runs_todo=new list<MANAGED_RUN*>;
  runs_inprogress=new list<MANAGED_RUN*>;
  clients_waiting=new multiset<CLIENT*,cmp_speed>;
  clients_working=new multiset<CLIENT*,cmp_speed>;

}

//****************************************************************************
void RUN_MANAGER::add_new_run(MODEL_RUN *_run)
//****************************************************************************
//CTM Apr 2011

/*
      Add new run to list of runs to be completed
*/

{
  MANAGED_RUN *newrun;

  newrun=new MANAGED_RUN;
  newrun->set_run(_run);
  newrun->set_status(MRUN_WAITING);
  runs_todo->push_back(newrun);
}

//****************************************************************************
void RUN_MANAGER::add_new_client(CLIENT *_slave)
//****************************************************************************
//CTM Apr 2011

/*
      Add new slave to list of slaves waiting to execute runs
*/

{

  clients_waiting->insert(_slave);
  _slave->status=CLIENT_WAITING;

}

//****************************************************************************
void RUN_MANAGER::start_waiting_runs()
//****************************************************************************
//CTM Apr 2011

/*
      Start any runs in the todo queue on available slaves
*/

{

  list<MANAGED_RUN*>::iterator run;
  multiset<CLIENT*,cmp_speed>::iterator client;

  // check if there are any runs to do
  if(runs_todo->size()>0)
  {
    client=clients_waiting->begin();
    run=runs_todo->begin();
    while(run!=runs_todo->end())
    {
      if(client==clients_waiting->end()) break;
      // give next available run to next available slave
      if(!(*run)->add_slave(*client))
        ++run;
      else
      {
        (*client)->status=CLIENT_WORKING;
        clients_working->insert(*client);
        clients_waiting->erase(client++);
        (*run)->set_status(MRUN_INPROGRESS);
        runs_inprogress->push_back(*run);
        run=runs_todo->erase(run);
      }
    }
  }

}

//****************************************************************************
void RUN_MANAGER::check_client_status(map<CLIENT*,bool> &_slaveconnected)
//****************************************************************************
//CTM Apr 2011

/*
      Check if any clients have disconnected
      If a client is the only one executing a run and it has disconnected
      then move that run back to the todo queue
*/

{

  list<MANAGED_RUN*>::iterator run;
  multiset<CLIENT*,cmp_speed>::iterator client;

  // check if the client(s) associated with any active runs have disconnected
  //cout<<"check if the client(s) associated with any active runs have disconnected"<<'\n';
  run=runs_inprogress->begin();
  while(run!=runs_inprogress->end())
  {
    (*run)->check_client_status();
    if((*run)->client_count()==0)
    {
      (*run)->set_status(MRUN_WAITING);
      runs_todo->push_back(*run);
      run=runs_inprogress->erase(run);
    }
    else
      ++run;
  }

  // update the working/waiting queues and delete any disconnected clients
  //cout<<"update the working/waiting queues and delete any disconnected clients"<<'\n';
  //cout<<"number of waiting slaves: "<<clients_waiting->size()<<'\n';
  //cout<<"number of working slaves: "<<clients_working->size()<<'\n';
  _slaveconnected.clear();
  client=clients_waiting->begin();
  while(client!=clients_waiting->end())
  {
    if((*client)->status==CLIENT_LOST)
    {
      cout<<"1disconnecting slave: "<<(*client)->name<<'\n';
      _slaveconnected[*client]=false;
      delete *client;
      clients_waiting->erase(client++);
    }
    else
    {
      _slaveconnected[*client]=true;
      ++client;
    }
  }
  client=clients_working->begin();
  while(client!=clients_working->end())
  {
    if((*client)->status==CLIENT_LOST)
    {
      cout<<"2disconnecting slave: "<<(*client)->name<<'\n';
      _slaveconnected[*client]=false;
      delete *client;
      clients_working->erase(client++);
    }
    else
    {
      _slaveconnected[*client]=true;
      ++client;
    }
  }

}


//****************************************************************************
void RUN_MANAGER::check_run_status()
//****************************************************************************
//CTM Apr 2011

/*
      Flag any runs that have completed
*/

{

  list<MANAGED_RUN*>::iterator run;
  //list<MANAGED_RUN*>::iterator run_copy;

  run=runs_inprogress->begin();
  while(run!=runs_inprogress->end())
  {
    if((*run)->status==MRUN_COMPLETE)
    {
      //cout<<"Run is complete."<<'\n';
      // kill run on other slaves
      (*run)->kill_runs(clients_waiting,clients_working);
      // add run to completed queue
      //_completedruns->push_back((*run)->run);
      // delete managed run
      //run_copy=run;
      (*run)->run->dealloc();
      delete (*run)->run;
      delete *run;
      run=runs_inprogress->erase(run);
    }
    else
      ++run;
  }
  
}

//****************************************************************************
void RUN_MANAGER::terminate_clients()
//****************************************************************************
//CTM Apr 2011

/*
      Terminate connection with all slaves
*/

{

  multiset<CLIENT*,cmp_speed>::iterator client;

  for(client=clients_working->begin();client!=clients_working->end();++client)
  {
    cout<<"     - "<<(*client)->name<<'\n';
    if(!(*client)->sendcommand(KILL_SLAVE))
      cout<<"Error sending command to slave."<<'\n';
    (*client)->stop();
    delete *client;
  }
  clients_working->clear();
  delete clients_working;

  for(client=clients_waiting->begin();client!=clients_waiting->end();++client)
  {
    cout<<"     - "<<(*client)->name<<'\n';
    if(!(*client)->sendcommand(KILL_SLAVE))
      cout<<"Error sending command to slave."<<'\n';
    (*client)->stop();
    delete *client;
  }
  clients_waiting->clear();
  delete clients_waiting;

  // delete run queues
  runs_todo->clear();
  delete runs_todo;
  runs_inprogress->clear();
  delete runs_inprogress;

}